
#ifndef _win32_GB_emu_H_
#define _win32_GB_emu_H_

#include <windows.h>   // include important windows stuff
#include <windowsx.h>
#include <process.h>

#include "emulator.h"
//#include "gameboy.h"
#include "gnuboy.h"
#include "win32_GB_screen_mgr.h"
//#include "input_mgr.h"
#include "win32_directsound_sound_mgr.h"
#include "null_sound_mgr.h"
#include "GB_pad.h"
#include "win32_GB_pad.h"
#include "win32_EXT_key.h"
#include "win32mic.h"


class win32_GB_emu : public emulator
{
public:
	win32_GB_emu(HWND parent_window_handle, HINSTANCE parent_instance_handle, const char* ROM_name);
	~win32_GB_emu();

	void PollInput(float);

	// emulator interface
	const char* getROMname();
	const char* getROMpath();
	boolean loadState(const char* fn);
	boolean saveState(const char* fn);

	void writeBaram(unsigned char* adr);		//Cheat
	void readBaram(unsigned char* adr);
	void ReadBoneb(unsigned int addr, unsigned char *data);
	void WriteBoneb(unsigned int addr, unsigned char data);

	void writeSraram(unsigned char* adr);
	void readSraram(unsigned char* adr);
	void ReadSroneb(unsigned int addr, unsigned char *data);
	void WriteSroneb(unsigned int addr, unsigned char data);

	uint32 GetMainMemp(uint8 **p){ return emu->GetMainMemp(p); };
	uint32 GetSramMemp(uint8 **p){ return emu->GetSramMemp(p); };
	uint32 GetExtMemp(uint8 **p){ return emu->GetExtMemp(p); };
	uint32 Get_PPUMemp(uint8 **p){ return emu->Get_PPUMemp(p); };
	void  emulate_CPU_cycles(float num_cycles){ /*emu->emulate_CPU_cycles(num_cycles);*/ };

	void GetCPUContext(unsigned char *context){ emu->GetCPUContext(context); }
	void SetCPUContext(unsigned char *context){ /*emu->SetCPUContext(context);*/}

	int get_emutype(){ return EMUTYPE_GB; }


	void ExtKeyExe();
	unsigned char GetScreenMode();
	void SetScreenMode(uint8 mode){
		//emu->SetScreenMode(mode);
		SetFrameSpeed(); }
	void nnnKailleraClient();	//nnn
	void nnnManuEnable(unsigned char);
	void GetROMInfoStr(char *wt);
	uint32 crc32(){return emu->crc32();}
	void SetFrameSpeed();
	void Load_Genie(){emu->Load_Genie();}
	void GetGameTitleName(char *s){ emu->GetGameTitleName(s); }


	void reset(unsigned char);
	void freeze();
	void thaw();
	void do_frame();
	boolean frozen()  { return emu->frozen(); }

	// screen manager interface
	void blt();
	void flip();
	void assert_palette();
	//  boolean toggle_fullscreen();
	//  boolean is_fullscreen()      { return scr_mgr->is_fullscreen(); }

	// sound interface
	void enable_sound(boolean enable);
	boolean sound_enabled();
	boolean set_sample_rate(int sample_rate);
	int get_sample_rate();
	void sound_settings_changed();

	// called when input settings are changed
	void input_settings_changed();

	// output *.wav, written by Mikami Kana
	bool start_sndrec(char *wfn);
	void end_sndrec();
	bool IsRecording();
	void ToggleFastFPS();

	//for Disk System
	uint8 GetDiskSideNum();
	uint8 GetDiskSide();
	void SetDiskSide(uint8 side);
	uint8 DiskAccessed();

	// for Expand Controllers
	void SetExControllerType(uint8 num);
	uint8 GetExControllerType();

	// for Screen Shot /by mikami
	void shot_screen(char *);

	void SetBarcodeValue(uint32 value_low, uint32 value_high);
	void StopTape();
	void StartPlayTape(const char* fn);
	void StartRecTape(const char* fn);
	uint8 GetTapeStatus();

	// for Movie
	void StopMovie(){ emu->StopMovie();}
	void StartPlayMovie(const char* fn, unsigned char mflag){ /*emu->StartPlayMovie(fn, mflag); */}
	void StartRecMovie(const char* fn, unsigned char mflag){ /*emu->StartRecMovie(fn, mflag); */}
	void StartPlayToRecMovie(){ /*emu->StartPlayToRecMovie();*/ }
	uint8 GetMovieStatus(){ return 0; } //emu->GetMovieStatus();}
	void InsertMovieMsg(char *msg){ /*emu->InsertMovieMsg(msg); */}
	void DeleteMovieMsg(){ /*emu->DeleteMovieMsg(); */}
	void MovieRecPause(){ /*emu->MovieRecPause(); */}
	void MoviePreStateRec(){ /*emu->MoviePreStateRec(); */}

	//  void SetScreenmgr();		//
	int start_avirec(char *fn){ return scr_mgr->start_avirec(fn); }
	void end_avirec(){ scr_mgr->end_avirec(); }
	int is_avirec(){ return scr_mgr->is_avirec(); }

	bool IsUserPause(){ return UserPause; };
	void SetUserPause(bool sw){ UserPause = sw; };

protected:
	HWND parent_wnd_handle;

	win32_GB_screen_mgr* scr_mgr;
	win32_directinput_input_mgr* inp_mgr;
	sound_mgr *snd_mgr;
	GAMEBOY *emu;		//emulator* emu;
	GB_pad pad1;
	GB_pad pad2;
	GB_pad pad3;
	GB_pad pad4;
	unsigned char pad_bit[6];

	bool UserPause;

	// token, local null sound manager; always there
	null_sound_mgr local_null_snd_mgr;

	void CreateWin32Pads();
	void DeleteWin32Pads();
	win32_GB_pad* win32_pad1;
	win32_GB_pad* win32_pad2;
	win32_GB_pad* win32_pad3;
	win32_GB_pad* win32_pad4;
//	win32_NES_button* win32_mic;
//	win32_NES_button* win32_coin;

	void CreateWin32Extkey();
	void DeleteWin32Extkey();
	win32_EXT_key* win32_extkey;

	boolean emulate_frame(boolean draw);
//	boolean emulate_frame_debug(boolean draw);

	double frames_per_sec;
	double last_frame_time;
	double cur_time;


	// profiling vars
	uint32 frames_this_sec;
	double last_profile_sec_time;
	uint32 skipframes_this_sec;

	void reset_last_frame_time();

	double FramePeriod;
	BOOL IsFastFPS;

	BOOL recaviflag;

	//Movie
/*
	HANDLE hFile_Movie;
	uint8 movie_flag;
	uint8 movie_status;
*/
private:
};

#endif
